package VB::Project;

# TODO:
# 1. Better mechanism for getting vbp file from reference detail
# 2. Optionally update clsid and version of references in vbp
# 3. Bug in vb filter - reduces spaces inside quotes
# 4. A way to find project in first place so we can execute from anywhere
# 5. Error output that will be useful to vim
# 6. Eliminate extra pass for NeedsCompile (means better mechanism for
#                                           unloading dlls)
# 7. try SWIG?  XS?  -- library fcn can be get_next_line
#                       to provide filtered lines one at a time.
#                       would be even better I imagine to combine with md5
# 8. some way to track the vb6.exe cpu usage to make sure it's working

use strict;
use warnings;

use vars '$AUTOLOAD';
use Carp;
use Cwd qw(abs_path cwd);
use File::Basename;
use Filesys::CygwinPaths ':all';
use Storable;
use Fingerprint::VB;
use Fingerprint::TypeLib;
use File::Basename;
use VB::Code;
use VB::Compiler;

my $cyg = 1;

{
    my $_compile_count = 0;

    sub compileCount { $_compile_count;}
    sub _incr_compile_count { ++$_compile_count; }
}

sub new
{
    my ($class, $vbpFile) = @_;
    
    $vbpFile =~ s/\.?$/.vbp/ unless $vbpFile =~ /\.vbp$/i;
    $vbpFile = my_abs_path($vbpFile);

    open(FH, $vbpFile) or croak "Unable to open file $vbpFile: $!\n";
    my @vbpText = <FH>;
    close FH or croak "Error closing file $vbpFile: $!\n";

    my $dataFile = $vbpFile;
    $dataFile =~ s/vbp$/sig/;
    my $persistentData = -f $dataFile ? retrieve($dataFile) : undef;

    my $objRef = { _vbp => $vbpFile,
                   _text => \@vbpText,
                   _path => dirname($vbpFile) . "/",
                   _data => $persistentData,
                   _datafile => $dataFile };

    bless $objRef, ref($class) || $class;
    return $objRef;
}

sub Force
{
    my $self = shift;
    $self->{_force} = 1;
}

sub NeedsCompile
{
    my $self = shift;
    my $recurse = shift;

    if ($recurse)
    {
        foreach my $refProj ($self->ReferenceProjects)
        {
#           print $self->Name . " calling NeedsCompile(1) on $refProj\n";
            return 20 if VB::Project->new($refProj)->NeedsCompile(1);
        } 
    }

    my $exe  = $self->_path . $self->ExeName32;
#   print "Checking $exe\n";

    return 1 unless -s $exe;
    return 2 unless defined $self->{_data};
    return 3 if $self->_force;

    # current file list
    my %files = $self->Files;
    my @refs  = $self->References;

    # check for missing files
    my @missing = grep !-e $_, values(%files);
    carp("Missing Files:\n", join("\n",@missing)) and return 0 if @missing;

    my $fileprints = $self->{_data}->{files};

    # Check for changes in file list
    return 4 if join("\r", sort(keys(%files)))
                ne join("\r", sort(keys(%$fileprints)));
    

    my $libprints = $self->{_data}->{refs};

    # Check for changes in library references list
    return 5 if join("\r", sort(@refs))
                ne join("\r", sort(keys(%$libprints)));
    
    # Check for changes in project
    return 6 if join('', $self->Build) ne join('', @{$self->{_data}->{build}});

    # Check for individual file changes
    my ($file, $fingerprint);
    while (($file, $fingerprint) = each %$fileprints)
    {
        return 20 unless defined $fingerprint;
        return 7 unless $fingerprint->isSame($files{$file});
    }

    # Check for library interface changes
    my $ref;
    while (($ref, $fingerprint) = each %$libprints)
    {
        #print $self->Name, " checking fingerprint for $ref\n";
        return 20 unless defined $fingerprint;
        return 8 unless $fingerprint->isSame($ref);
    }

    return 0;
}

sub Compile
{
    my ($self, $recursion) = @_;

    foreach my $refProj ($self->ReferenceProjects)
    {
#       print $self->Name . " calling compile on $refProj\n";
        VB::Project->new($refProj)->Compile() or return 0;
    }

    my $needsCompile = $self->NeedsCompile;
#   print $self->_vbp . " - NeedsCompile = $needsCompile\n";
    return 1 unless $needsCompile;
    
    my $component = $self->_path . $self->ExeName32;
    #unload($component);

    print "compiling ", $self->Name . "\n" unless $recursion;

    my $project  = $self->_vbp;

    my @output = VB::Compiler->new->Make($project);

    if ($output[-1] =~ /Build of '.*' succeeded./)
    {
        print $output[-1];
        _incr_compile_count();
        $self->StoreFingerprint();
        return 1;
    }
    elsif ($output[1] =~ /(binary|version)[- ]compatib/
            and $self->GetCompatibleMode() == 2 
            and !$recursion)
    {
        # keep track of recursion, just in case.  It's a messy business

        # determine what compatibility level to use
        my $levelNum = $output[1] =~ /version compatible/ ? 0 : 1;
        my $levelStr = $levelNum == 1 ? "Project" : "No";
        print "compatibility error, attempting with $levelStr Compatibility ...\n";

        $self->SetCompatibleMode($levelNum);
        my $result = $self->Compile(1);
        $self->SetCompatibleMode(2);

        return $result;
    }
    else
    {
        #my @fileErrors = grep /in File '.*', Line \d/, @output;
        #displayErrors(@fileErrors) if @fileErrors;
        # displayErrors already filters output
        displayErrors(@output); 
        die @output;
        #return 0;
    }
}

sub GetCompatibleMode
{
    my $self = shift;
    for my $line ($self->_text)
    {
        return $1 if $line =~ /^CompatibleMode="(\d)/;
    }
}

sub SetCompatibleMode
{
    my ($self, $compat) = @_;

    open PFH, ">".$self->_vbp() or croak "unable to open project file: $!\n";
    foreach my $line ($self->_text)
    {
        $line =~ s/^(CompatibleMode=")\d\"/$1$compat"/;
        print PFH $line;
    }

    close PFH or croak "unable to close project file: $!\n";
}

sub displayErrors
{
    ERRORMSG: foreach my $errorMessage (@_)
    {
        next ERRORMSG unless $errorMessage =~ /in File '(.*)', Line (\d+) : (.*)/;
        my $file = $1;
        my $line = $2;
        my $message = $3;

        fixSlashes($file);
        my $margin = 1 + length($line + 5);

        my $code = VB::Code->new($file);
        CODELINE: for (my $i = $line - 5; $i <= $line + 5; $i++)
        {
            # make sure we're not out of bounds
            next CODELINE if $i < 0;
            last CODELINE unless defined $code->Code($i);

            print $i == $line ? ">" : " ";
            printf "%${margin}d  ", $i;
            print $code->Code($i) ;
        }
    }
}

sub StoreFingerprint
{
    my $self = shift;
    my %files = $self->Files;
    my @build = $self->Build;

    my $oldprints;
    my $newprints;

#   print("Storing fingerprint for " . $self->Name . " build\n");

    if (defined $self->{_data} and ref $self->{_data}->{files} eq 'HASH')
    {
        $oldprints = $self->{_data}->{files};
        foreach my $file (keys %files)
        {
            if (defined $oldprints->{$file})
            {
                $newprints->{$file} = $oldprints->{$file};
                $newprints->{$file}->update();
            }
            else
            {
                $newprints->{$file} = Fingerprint::VB->new($files{$file})->cache;
            }
        }
    }
    else
    {
        $newprints = {};
        foreach my $file (keys %files)
        {
            $newprints->{$file} = Fingerprint::VB->new($files{$file})->cache;
        }
    }

    my %refs;
    foreach my $ref ($self->References)
    {
        $refs{$ref} = Fingerprint::TypeLib->new($ref);
    }

    $self->{_data} = {files => $newprints, build => \@build, refs => \%refs};

    store $self->{_data}, $self->{_datafile};
}

sub ReferenceProjects
{
    my $self = shift;
    # don't cache these and don't instantiate VB::Project objects yet
    # since there might be diamond-dependencies that are compiled and
    # revisited.
    #return $self->_projs if defined $self->{_projs};

    my @projects;

    my @deps = $self->References;
    map s/\.\w+$/.vbp/i, @deps;
    return (grep -e $_, @deps);

    #$self->{_projs} = \@projects;
    #return @projects;
}

sub References
{
    my $self = shift;
    # careful!  $self->_refs returns a hash (array)
    #           $self->{_refs} is a hash ref
    return $self->_refs if defined $self->{_refs};

    my $olddir = cwd;
    chdir $self->_path;

    my @refs;
    foreach my $line (grep /^Reference=/, $self->_text)
    {
        my $path = (split /#/, $line)[3];
        
        push @refs, my_lib_path($path);
    }

    chdir $olddir;

    $self->{_refs} = \@refs;
    return @refs;
}

sub my_lib_path
{
    my $lib = shift;
    my ($path, $ext) = $lib =~ /^(.*\.(?:dll|tlb|ocx))(\\\d+)?$/i;
    return "" unless $path;
    $ext ||= "";
    return my_abs_path($path) . $ext;
}

sub Files
{
    my $self = shift;
    # careful!  $self->_files returns a hash (array)
    #           $self->{_files} is a hash ref
    return $self->_files if defined $self->{_files};

    my %files;
    my $olddir = cwd;
    chdir $self->_path;

    foreach my $line ($self->_text)
    {
        if ($line =~ /^(?:Class|Module)=(\w+); (.+)\r\n/)
        {
            $files{$1} = my_abs_path($2);
        }
    }

    chdir $olddir;

    $self->{_files} = \%files;
    return %files;
}

sub Build
{
    my $self = shift;
    # careful!  $self->_build returns a hash (array)
    #           $self->{_build} is a hash ref
    return $self->_build if defined $self->{_build};
    
    my @build = grep !/^(?:Class|Module|Reference)=/, $self->_text;

    $self->{_build} = \@build;
    return @build;
}

sub my_abs_path
{
    my $path = shift;
    if ($cyg)
    {
        $path = fullwin32path($path);
        fixSlashes($path);
    }
    else
    {
        $path = abs_path($path);
    }
    return $path;
}

sub fixSlashes
{
    map s"\\"/"g, @_;
}

sub AUTOLOAD
{
    my $self = shift;
    my $class = ref $self;
    if ($AUTOLOAD =~ /${class}::(Type|Name|ExeName32|Description)/)
    {
        my $prop = $1;
        return $self->{"_$prop"} if defined $self->{"_$prop"};

        my ($dataline) = grep /^$prop="(.*)"/, $self->_text;
        return "" unless defined $dataline;
        $dataline =~ /$prop="(.*)"/;
        $self->{"_$prop"} = $1;
        return $self->{"_$prop"};
    }

    if ($AUTOLOAD =~ /::(_\w+)$/)
    {
        my $ptr = $self->{$1};
        return '' unless defined $ptr;

        my $type = ref $ptr;
        if    ($type eq 'SCALAR') { return ${$ptr} }
        elsif ($type eq 'ARRAY')  { return @{$ptr} }
        elsif ($type eq 'HASH')   { return %{$ptr} }
        else                      { return $ptr    }
    }

    croak "Method $AUTOLOAD does not exist";
}

sub DESTROY {}

1;
